﻿using System;
using Cysharp.Threading.Tasks;
using UniRx;
using UnityEngine;

namespace Magify
{
    internal static class ApplicationHelper
    {
        public enum RestoreType
        {
            Unsupported,
            ForegroundEvent,
            Timer,
        }

        private static IObservable<bool> _everyApplicationFocus;

        public static bool IsEditor(this RuntimePlatform platform)
        {
            return platform is RuntimePlatform.WindowsEditor or RuntimePlatform.LinuxEditor or RuntimePlatform.OSXEditor;
        }

        public static async UniTask<RestoreType> CallAndWaitForApplicationRestore(Action method)
        {
            if (method == null)
            {
                throw new ArgumentNullException(nameof(method));
            }

            // This method well tested only for Android, iOS and Editor. Ignore awaiting focus on other platforms
            if (!Application.isEditor && Application.platform != RuntimePlatform.Android && Application.platform != RuntimePlatform.IPhonePlayer)
            {
                method();
                return RestoreType.Unsupported;
            }

            using var disposer = new CompositeDisposable();
            var source = new UniTaskCompletionSource<RestoreType>();

            EveryApplicationFocus()
                .Where(c => c)
                .ObserveOnMainThread()
                .First()
                .Subscribe(_ => { source.TrySetResult(RestoreType.ForegroundEvent); })
                .AddTo(disposer);

            // In some cases system does not send focus and pause events correctly.
            // For that matter finish awaiting after half second
            if (!Application.isEditor)
            {
                Observable.Timer(TimeSpan.FromSeconds(0.5))
                    .First()
                    .Subscribe(_ => { source.TrySetResult(RestoreType.Timer); })
                    .AddTo(disposer);
            }

            method();

            return await source.Task;
        }

        /// <summary>
        /// Same as Observable.EveryApplicationFocus(), but is not invoking when focus is lost on the NotificationCenter shown.
        /// Use ObserveOnMainThread() if you want to handle it on main thread
        /// </summary>
        public static IObservable<bool> EveryApplicationFocus()
        {
            switch (Application.platform)
            {
                case RuntimePlatform.Android or RuntimePlatform.IPhonePlayer when !Application.isEditor && MagifyManager.Initialized:
                    if (_everyApplicationFocus == null)
                    {
                        var observableForeground = Observable.FromEvent(t => MagifyManager.OnApplicationEnterForeground += t, t => MagifyManager.OnApplicationEnterForeground -= t)
                            .Select(_ => true);
                        var observableBackground = Observable.FromEvent(t => MagifyManager.OnApplicationEnterBackground += t, t => MagifyManager.OnApplicationEnterBackground -= t)
                            .Select(_ => false);

                        _everyApplicationFocus = observableForeground.Merge(observableBackground);
                    }
                    return _everyApplicationFocus;
                default:
                    return Observable.EveryApplicationFocus();
            }
        }
    }
}